@@ -26,11 +26,21 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
   end
 
   def warning(str)
-    parse(str, :warning) {|e, *args| return args}
+    tree = parse(str, :warning) {|e, *args| return args}
+    if block_given?
+      yield tree
+    else
+      assert(false, "warning expected")
+    end
   end
 
   def warn(str)
-    parse(str, :warn) {|e, *args| return args}
+    tree = parse(str, :warn) {|e, *args| return args}
+    if block_given?
+      yield tree
+    else
+      assert(false, "warning expected")
+    end
   end
 
   def test_program
@@ -132,9 +142,17 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
   end
 
   def test_args_forward
-    thru_args_forward = false
-    parse('def m(...) n(...) end', :on_args_forward) {thru_args_forward = true}
-    assert_equal true, thru_args_forward
+    [
+      'def m(...) n(...) end',
+      'def m(...) end',
+      'def m(a, ...) n(1, ...) end',
+      'def m(...) n(1, ...) end',
+      'def m(a, ...) n(...) end'
+    ].each do |code|
+      thru_args_forward = false
+      parse(code, :on_args_forward) {thru_args_forward = true}
+      assert_equal true, thru_args_forward, "no args_forward for: #{code}"
+    end
   end
 
   def test_arg_paren
@@ -443,13 +461,6 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
     assert_equal "[call(ref(self),&.,foo,[])]", tree
   end
 
-  def test_methref
-    thru_methref = false
-    tree = parse("obj.:foo", :on_methref) {thru_methref = true}
-    assert_equal true, thru_methref
-    assert_equal "[methref(vcall(obj),foo)]", tree
-  end
-
   def test_excessed_comma
     thru_excessed_comma = false
     parse("proc{|x,|}", :on_excessed_comma) {thru_excessed_comma = true}
@@ -651,6 +662,30 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
     assert_equal '[def(foo,[],bodystmt([void()]))]', parse('def foo ;end')
   end
 
+  def test_endless_def
+    events = %i[on_def on_parse_error]
+    thru = nil
+    hook = ->(name, *) {thru[name] = true}
+
+    thru = {}
+    tree = parse('def foo() = 42', events, &hook)
+    assert_equal({on_def: true}, thru)
+    assert_equal '[def(foo,[],42)]', tree
+
+    thru = {}
+    tree = parse('def foo() = 42 rescue 0', events, &hook)
+    assert_equal({on_def: true}, thru)
+    assert_equal '[def(foo,[],rescue_mod(42,0))]', tree
+
+    thru = {}
+    tree = parse('def foo=() = 42', events, &hook)
+    assert_equal({on_def: true, on_parse_error: true}, thru)
+
+    thru = {}
+    tree = parse('def foo=() = 42 rescue 0', events, &hook)
+    assert_equal({on_def: true, on_parse_error: true}, thru)
+  end
+
   def test_defined
     thru_defined = false
     parse('defined?(x)', :on_defined) {thru_defined = true}
@@ -668,6 +703,30 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
     assert_equal(true, thru_parse_error)
   end
 
+  def test_endless_defs
+    events = %i[on_defs on_parse_error]
+    thru = nil
+    hook = ->(name, *) {thru[name] = true}
+
+    thru = {}
+    tree = parse('def foo.bar() = 42', events, &hook)
+    assert_equal({on_defs: true}, thru)
+    assert_equal '[defs(vcall(foo),.,bar,[],42)]', tree
+
+    thru = {}
+    tree = parse('def foo.bar() = 42 rescue 0', events, &hook)
+    assert_equal({on_defs: true}, thru)
+    assert_equal '[defs(vcall(foo),.,bar,[],rescue_mod(42,0))]', tree
+
+    thru = {}
+    tree = parse('def foo.bar=() = 42', events, &hook)
+    assert_equal({on_defs: true, on_parse_error: true}, thru)
+
+    thru = {}
+    tree = parse('def foo.bar=() = 42 rescue 0', events, &hook)
+    assert_equal({on_defs: true, on_parse_error: true}, thru)
+  end
+
   def test_do_block
     thru_do_block = false
     parse('proc do end', :on_do_block) {thru_do_block = true}
@@ -1497,16 +1556,9 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
   end
 
   def test_block_variables
-    assert_equal("[fcall(proc,[],&block([],[void()]))]", parse("proc{|;y|}"))
-    if defined?(Process::RLIMIT_AS)
-      dir = File.dirname(__FILE__)
-      as = 100 * 1024 * 1024 # 100MB
-      as *= 2 if RubyVM::MJIT.enabled? # space for compiler
-      assert_in_out_err(%W(-I#{dir} -rdummyparser),
-                        "Process.setrlimit(Process::RLIMIT_AS,#{as}); "\
-                        "puts DummyParser.new('proc{|;y|!y}').parse",
-                        ["[fcall(proc,[],&block([],[unary(!,ref(y))]))]"], [], '[ruby-dev:39423]')
-    end
+    bug4159 = '[ruby-dev:39423]'
+    assert_equal("[fcall(proc,[],&block([],[void()]))]", parse("proc{|;y|}"), bug4159)
+    assert_equal("[fcall(proc,[],&block([],[unary(!,ref(y))]))]", parse("proc{|;y|!y}"), bug4159)
   end
 
   def test_unterminated_regexp
@@ -1536,12 +1588,28 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
     assert_equal("frozen_string_literal", args[0])
   end
 
+  def test_warning_invalid_magic_comment
+    fmt, *args = warning("#-*- frozen-string-literal: nottrue -*-")
+    assert_match(/invalid value/, fmt)
+    assert_equal(%w"frozen_string_literal nottrue", args)
+  end
+
   def test_warn_cr_in_middle
     fmt = nil
     assert_warn("") {fmt, = warn("\r;")}
     assert_match(/encountered/, fmt)
   end
 
+  def test_warn_mismatched_indentations
+    fmt, tokend, tokbeg, line = assert_warning("") {break warn("if true\n  end\n")}
+    assert_match(/mismatched indentations/, fmt)
+    assert_equal(["if", "end", 1], [tokbeg, tokend, line])
+    result = assert_warning("") {
+      warn("begin\n" "  def f() = nil\n" "end\n") {break :ok}
+    }
+    assert_equal(:ok, result)
+  end
+
   def test_in
     thru_in = false
     parse('case 0; in 0; end', :on_in) {thru_in = true}
@@ -1554,6 +1622,12 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
     assert_equal true, thru_aryptn
   end
 
+  def test_fndptn
+    thru_fndptn = false
+    parse('case 0; in [*,0,*]; end', :on_fndptn) {thru_fndptn = true}
+    assert_equal true, thru_fndptn
+  end
+
   def test_hshptn
     thru_hshptn = false
     parse('case 0; in {a:}; end', :on_hshptn) {thru_hshptn = true}
